package geniusweb.protocol.session.mopac;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.mockito.Matchers.any;
import static org.mockito.Matchers.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import java.io.IOException;
import java.net.URISyntaxException;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Map;

import org.junit.Before;
import org.junit.Test;

import com.fasterxml.jackson.core.JsonParseException;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;

import geniusweb.actions.Action;
import geniusweb.actions.PartyId;
import geniusweb.progress.Progress;
import geniusweb.progress.ProgressTime;
import geniusweb.protocol.ProtocolException;
import geniusweb.protocol.session.OnePartyTeam;
import geniusweb.protocol.session.mopac.phase.OfferPhase;
import geniusweb.protocol.session.mopac.phase.Phase;
import geniusweb.references.Parameters;
import geniusweb.references.PartyWithParameters;
import geniusweb.references.PartyWithProfile;
import geniusweb.references.ProfileRef;
import geniusweb.voting.votingevaluators.LargestAgreement;
import tudelft.utilities.junit.GeneralTests;

public class MOPACStateTest extends GeneralTests<MOPACState> {
	private static final Date NOW = new Date(1001);

	private final ObjectMapper jackson = new ObjectMapper();

	private final MOPACSettings settings = mock(MOPACSettings.class);

	private final MOPACState state = new MOPACState(settings);

	private PartyId PARTY1 = new PartyId("party1");
	private PartyId PARTY2 = new PartyId("party2");
	private PartyId PARTY3 = new PartyId("party3");

	private PartyWithParameters partyparams1 = mock(PartyWithParameters.class);
	private Parameters params1 = new Parameters();

	private PartyWithProfile profile1 = new PartyWithProfile(partyparams1,
			mock(ProfileRef.class));

	private final String mopacsettingstr = "{\"MOPACSettings\":{\"participants\":[{\"party\":{\"partyref\":\"http://party1\",\"parameters\":{}},\"profile\":\"http://profile1\"},{\"party\":{\"partyref\":\"http://party2\",\"parameters\":{}},\"profile\":\"http://profile2\"}],\"deadline\":{\"deadlinetime\":{\"durationms\":100}},\"votingevaluator\":{\"LargestAgreement\":{}}}}";
	// create real objects for the GeneralTests and for serialization tests
	private PartyStates partyStates = new PartyStates(Collections.emptyMap());
	private Phase phase1 = new OfferPhase(partyStates, 100l,
			new LargestAgreement());
	private Phase phase2 = new OfferPhase(partyStates, 200l,
			new LargestAgreement());
	private Progress progress1 = new ProgressTime(100l, NOW);
	private Progress progress2 = new ProgressTime(200l, NOW);

	private MOPACState state1, state1a, state2, state3, state4, state5, state6;

	private String serialized = "{\"MOPACState\":{\"phase\":{\"OfferPhase\":{\"partyStates\":{\"notYetActed\":[],\"actions\":[],\"agreements\":{},\"walkedAway\":[],\"exceptions\":{},\"powers\":{}},\"evaluator\":{\"LargestAgreement\":{}},\"deadline\":100}},\"actions\":[],\"progress\":{\"time\":{\"duration\":100,\"start\":1001}},\"settings\":{\"MOPACSettings\":{\"participants\":[{\"party\":{\"partyref\":\"http://party1\",\"parameters\":{}},\"profile\":\"http://profile1\"},{\"party\":{\"partyref\":\"http://party2\",\"parameters\":{}},\"profile\":\"http://profile2\"}],\"deadline\":{\"deadlinetime\":{\"durationms\":100}},\"votingevaluator\":{\"LargestAgreement\":{}}}},\"partyprofiles\":{}}}";

	@Before
	public void before()
			throws JsonParseException, JsonMappingException, IOException {
		params1.put("power", 2);
		when(partyparams1.getParameters()).thenReturn(params1);

		MOPACSettings realsettings = jackson.readValue(mopacsettingstr,
				MOPACSettings.class);
		OnePartyTeam party = mock(OnePartyTeam.class);
		PartyWithParameters pwithp = mock(PartyWithParameters.class);
		when(pwithp.getParameters()).thenReturn(new Parameters());
		when(party.getParty()).thenReturn(pwithp);
		MOPACSettings realsettings2 = realsettings.with(party);

		state1 = new MOPACState(phase1, Collections.emptyList(), progress1,
				realsettings, Collections.emptyMap());
		state1a = new MOPACState(phase1, Collections.emptyList(), progress1,
				realsettings, Collections.emptyMap());
		state2 = new MOPACState(phase2, Collections.emptyList(), progress1,
				realsettings, Collections.emptyMap());
		state3 = new MOPACState(phase1, Arrays.asList(mock(Action.class)),
				progress1, realsettings, Collections.emptyMap());
		state4 = new MOPACState(phase1, Collections.emptyList(), progress2,
				realsettings, Collections.emptyMap());
		state5 = new MOPACState(phase1, Collections.emptyList(), progress1,
				realsettings2, Collections.emptyMap());
		state6 = new MOPACState(phase1, Collections.emptyList(), progress1,
				realsettings,
				Collections.singletonMap(PARTY1, mock(PartyWithProfile.class)));

	}

	@Override
	public List<List<MOPACState>> getGeneralTestData() {
		return Arrays.asList(Arrays.asList(state1, state1a),
				Arrays.asList(state2), Arrays.asList(state3),
				Arrays.asList(state4), Arrays.asList(state5),
				Arrays.asList(state6));
	}

	@Override
	public List<String> getGeneralTestStrings() {
		return Arrays.asList(
				"MOPACState.*OfferPhase.*PartyStates.*,MOPACSettings.*PartyWithProfile.*ProgressTime.*",
				"MOPACState.*OfferPhase.*PartyStates.*,MOPACSettings.*PartyWithProfile.*ProgressTime.*",
				"MOPACState.*OfferPhase.*PartyStates.*,MOPACSettings.*PartyWithProfile.*ProgressTime.*",
				"MOPACState.*OfferPhase.*PartyStates.*,MOPACSettings.*PartyWithProfile.*ProgressTime.*",
				"MOPACState.*OfferPhase.*PartyStates.*,MOPACSettings.*PartyWithProfile.*ProgressTime.*",
				"MOPACState.*OfferPhase.*PartyStates.*,MOPACSettings.*PartyWithProfile.*ProgressTime.*");
	}

	@Test
	public void smokeTest() {
	}

	@Test
	public void addProfilesBeforeInit() {
		MOPACState st = state.with(PARTY1, profile1);
		assertTrue(st.getPartyProfiles().containsKey(PARTY1));
	}

	@Test
	public void testInit() {
		// no parties were added, should throw? Go to final mode?
		// Notice that it's possible to create settings with 0 participants,
		// because this is needed for Tournament mechanism.
		MOPACState state1 = state.initPhase(new ProgressTime(100l, new Date(0)),
				1);
		assertTrue(state1.getActions().isEmpty());
		assertTrue(state1.getAgreements().getMap().isEmpty());
		assertTrue(state1.getPartyProfiles().isEmpty());
		Phase phase1 = state1.getPhase();
		assertTrue(phase1 instanceof OfferPhase);
		OfferPhase offer1 = (OfferPhase) phase1;
		assertEquals(new PartyStates(Collections.emptyMap()),
				offer1.getPartyStates());
	}

	@Test
	public void getAgreementsTest() {
		MOPACState st = state.initPhase(new ProgressTime(100l, new Date(0)), 1);
		assertTrue(st.getAgreements().getMap().isEmpty());
	}

	@Test
	public void testDeserialize() throws IOException {
		MOPACState obj = jackson.readValue(serialized, MOPACState.class);
		System.out.println(obj);
		assertEquals(state1, obj);
	}

	@Test
	public void testSerialize()
			throws JsonProcessingException, URISyntaxException {

		String string = jackson.writeValueAsString(state1);
		System.out.println(string);
		assertEquals(serialized, string);
	}

	@Test
	public void testNewPhasePossible() {

		Phase phase = mock(Phase.class);
		List<Action> actions = Collections.emptyList();
		Progress progress = mock(Progress.class);
		when(progress.getTerminationTime())
				.thenReturn(new Date(Phase.PHASE_MAXTIME)); // plenty for new
															// phase
		Map<PartyId, PartyWithProfile> partyprofiles = mock(Map.class);
		MOPACState s1 = new MOPACState(phase, actions, progress, settings,
				partyprofiles);

		PartyStates states = mock(PartyStates.class);
		when(phase.getPartyStates()).thenReturn(states);
		when(states.getNegotiatingParties())
				.thenReturn(new HashSet<>(Arrays.asList(PARTY1, PARTY2)));
		assertTrue(s1.isNewPhasePossible(10));
	}

	@Test
	public void testNewPhasePossibleOnly1Party() {

		Phase phase = mock(Phase.class);
		List<Action> actions = Collections.emptyList();
		Progress progress = mock(Progress.class);
		when(progress.getTerminationTime())
				.thenReturn(new Date(Phase.PHASE_MAXTIME)); // plenty for new
															// phase
		Map<PartyId, PartyWithProfile> partyprofiles = mock(Map.class);
		MOPACState s1 = new MOPACState(phase, actions, progress, settings,
				partyprofiles);

		PartyStates states = mock(PartyStates.class);
		when(phase.getPartyStates()).thenReturn(states);
		when(states.getNegotiatingParties())
				.thenReturn(new HashSet<>(Arrays.asList(PARTY1)));
		assertFalse(s1.isNewPhasePossible(10));
	}

	@Test
	public void testNewPhasePossibleNotEnoughTime() {

		Phase phase = mock(Phase.class);
		List<Action> actions = Collections.emptyList();
		Progress progress = mock(Progress.class);
		when(progress.getTerminationTime())
				.thenReturn(new Date(Phase.PHASE_MINTIME - 10));
		Map<PartyId, PartyWithProfile> partyprofiles = mock(Map.class);
		MOPACState s1 = new MOPACState(phase, actions, progress, settings,
				partyprofiles);

		PartyStates states = mock(PartyStates.class);
		when(phase.getPartyStates()).thenReturn(states);
		when(states.getNegotiatingParties())
				.thenReturn(new HashSet<>(Arrays.asList(PARTY1)));
		assertFalse(s1.isNewPhasePossible(10));
	}

	@Test
	public void nextPhaseTest() {
		long NOW = 102;
		Phase phase = mock(Phase.class);
		List<Action> actions = Collections.emptyList();
		Progress progress = mock(Progress.class);
		when(progress.getTerminationTime())
				.thenReturn(new Date(Phase.PHASE_MAXTIME));
		Map<PartyId, PartyWithProfile> partyprofiles = mock(Map.class);
		MOPACState s1 = new MOPACState(phase, actions, progress, settings,
				partyprofiles);

		// PhaseB is the next phase, with remaining time
		Phase phaseB = mock(Phase.class);
		when(phase.next(eq(NOW), eq(Phase.PHASE_MAXTIME - NOW)))
				.thenReturn(phaseB);

		MOPACState s2 = s1.nextPhase(NOW);
		assertEquals(phaseB, s2.getPhase());
	}

	@Test
	public void exceptionTest() {
		Phase phase = mock(Phase.class);
		List<Action> actions = Collections.emptyList();
		Progress progress = mock(Progress.class);
		Map<PartyId, PartyWithProfile> partyprofiles = mock(Map.class);
		MOPACState s1 = new MOPACState(phase, actions, progress, settings,
				partyprofiles);

		Phase exphase = mock(Phase.class); // supposed phase that contains the
											// exception
		when(phase.with(any(ProtocolException.class))).thenReturn(exphase);

		MOPACState errstate = s1.with(new ProtocolException("test", PARTY1));
		assertEquals(exphase, errstate.getPhase());
		assertEquals(progress, errstate.getProgress());

	}

}
